/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
`include "fifo_rtl.sv"
`include "chk_fifo.sv"
`include "cache_mem.sv"
module cache_controller(
                        //user interface
                        output logic data_vld, busy,  
                        input  logic rd, wr,
                        input logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr,                     
                        output logic [MEM_DATA_WIDTH-1:0]         rd_data, 
                        input logic  [MEM_DATA_WIDTH-1:0]         wr_data,  
                        // main memory interface 
                        output logic mem_rd, mem_wr, 
                        output logic [MAIN_MEM_ADDRESS_WIDTH-1:0] mem_addr, 
                        output logic [MEM_DATA_WIDTH-1:0]         mem_wr_data,  
                        input logic  [MEM_DATA_WIDTH-1:0]         mem_rd_data, 
                        // Signals needed by checker                        
                        output logic cache_exists,  // exists at specified address
                        output logic [MEM_DATA_WIDTH-1:0]         cache_data,

                        // clock and resets 
                        input logic clk, 
                        input logic rst_n);
  
  // Local declarations
  timeunit 1ns;     timeprecision 100ps;
  // ***** 
  chk_rd_wr_cntlr chk_rd_wr_cntlr_1(.*);
  
  chk_wr_cntrl chk_wr_cntrl_1(.*,
	  .cache_valid(cache_mem1.cache_valid), 
	  .cache_mem_data(cache_mem1.cache_mem_data), // cache data storage 
      .cache_mem_addr(cache_mem1.cache_mem_addr),  
      .cache_entry(cache_mem1.cache_entry));
  
  chk_reset chk_reset_1(.*);
  
  chk_rd_cntrl chk_rd_cntrl_1(.*, .cache_valid(cache_mem1.cache_valid));

  //  internal  controller signals
  logic[2:0]   wr_cycle_timer;  // write cycle timer 
  logic[2:0]   rd_cycle_timer;  // read cycle timer
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr_r1, addr_r2, addr_r3; 
  logic                              wr_cache;  // instructs cache to write a line 
  logic                              invalidate; // instructs cache to invalidate a line 
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] wr_addr;  // address to write into cache 
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr2invalidate; // address to invalidate
  logic [MEM_DATA_WIDTH-1:0]         cache_wr_data; // data to write to cache  
  logic                              check_cache;  // check the cache for exists
  logic                              data_from_mem;
  
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr_r; // delayed address from user interface
  
  typedef   logic [MAIN_MEM_ADDRESS_WIDTH-1 : 0] word_t;
  logic                                          miss;
  logic empty, full, push, pop;
  word_t last_entry; // last entry of addr into the fifo. Used when cache
  // reaches maximum size
  logic error; // fifo

 default clocking default_clk @ (posedge clk); endclocking
 default disable iff !rst_n;
  assign miss            = !cache_exists;
  assign check_cache = rd || wr;

 
  always @ (posedge clk) begin : filling_cache_invalidates
    addr_r              <= addr; // The addr_r is stored into the fifo
    // defaults assignments 
    pop <= 1'b0;
    push <= 1'b0;
    invalidate <= 1'b0;
	//wr_cache <= 1'b0;
	if((rd || wr) && miss) begin : read_write_miss
	  push    <= 1'b1;  // push the new entry (i.e., the new addr)
	  if(full) begin : rd_wr_full
         pop     <= 1'b1;  // pop the last entry
		 addr2invalidate       <= last_entry;
         invalidate            <= 1'b1;
	  end : rd_wr_full
	  // else if (wr && cache_exists) begin : wr_cache_exists
		  
	 end :read_write_miss
  end : filling_cache_invalidates

  
  // ********** MAIN MEMORY WRITES CONTROL **********
  // Main memory is slow, and requires 3 cycles to read or to write.  
  // Main memory size is parameterized.  
  // Data is latched on the falling edge of the write signal and is driven 
  // on the rising edge of the read signal.  
  // Handle the main write cycles 
    always @ (posedge clk) begin : main_mem_accesses 
      // defaults
      mem_rd   <= 1'b0;
      busy     <= 1'b0;
      data_vld <= 1'b0;
      mem_rd   <= 1'b0;
      mem_wr   <= 1'b0;
	  wr_cache <= 1'b0;
      // WRITES 
      if (wr) begin : wr_2main // Start of a write
		chk_immd chk_immd_wr_cycle_timer_not_0(
             .the_what(wr_cycle_timer==0),
             .msg("wr signal when memory counter !=0"),
             .clk(clk));

        wr_cycle_timer <= MEM_CYCLES-1'b1; // load cycle timer for latency 
        mem_wr_data       <= wr_data;
        mem_addr          <= addr;
        mem_wr            <= 1'b1;
        busy              <= 1'b1;
        // cache[mem_addr]    = wr_data; // fill the cache
		wr_cache <= 1'b1;
        cache_wr_data <= wr_data;
        wr_addr <= addr; 
      end : wr_2main
      if (wr_cycle_timer!=0) begin : dec_wr_cycle_timer 
        wr_cycle_timer <= wr_cycle_timer - 1'b1;
        mem_wr            <= 1'b1;
        busy              <= 1'b1;
      end  : dec_wr_cycle_timer
      // READS 
      if(rd && !miss) begin : rd_from_cache // must provide the data_vld
        data_vld           <= 1'b1;
      end : rd_from_cache // bagath miss 
      if(rd && miss) begin : rd_from_main
		chk_immd chk_immd_rd_cycle_timer_not_zero(
             .the_what(rd_cycle_timer==0),
             .msg("wr signal when memory counter !=0"),
             .clk(clk));

        rd_cycle_timer <= MEM_CYCLES - 1'b1; // load cycle timer for latency 
        mem_addr          <= addr;
        mem_rd            <= 1'b1;
        busy              <= 1'b1;
      end : rd_from_main
      if(rd_cycle_timer!=0) begin : rd_mem_in_progress
        rd_cycle_timer <= rd_cycle_timer - 1'b1; 
        mem_rd            <= 1'b1;
        busy              <= 1'b1;
      end : rd_mem_in_progress
      if(rd_cycle_timer==1) begin : write_to_cache 
        // cache[mem_addr]    = mem_rd_data;
		wr_cache <= 1'b1;
        cache_wr_data <= mem_rd_data;
        wr_addr <= mem_addr; // addr; 
        data_vld          <= 1'b1;    //bagath
        // rd_cycle_timer <= rd_cycle_timer - 1'b1; 
      end : write_to_cache
      if (!rst_n) begin : resets
        rd_cycle_timer <= 0;
        wr_cycle_timer <= 0;
        mem_rd   <= 1'b0;
        busy     <= 1'b0;
        data_vld <= 1'b0;
        mem_rd   <= 1'b0;
        mem_wr   <= 1'b0;
		wr_cache <= 1'b0;
        //cache.delete; // Clears all elements of associative array       
      end : resets      
    end : main_mem_accesses  
  
  

  /********** Read data to user interface **********/
  // Need the delayed signal // Below is in error 
  // assign rd_data = miss? mem_rd_data :  // If miss, return memory data
  //                 cache_data;    // If !miss, return cache data
  always @ (posedge clk) begin : rd_data_output
	if (rd && miss) begin : rd_miss_data_from_mem
      data_from_mem <= 1'b1;
	end :rd_miss_data_from_mem
	else if (rd && cache_exists) begin : rd_hit
      data_from_mem <= 1'b0;	
    end : rd_hit
	  end : rd_data_output

  assign rd_data = data_from_mem && data_vld ? mem_rd_data :  // If miss, return memory data
                  cache_data; // cache[addr];   // If hit, return cache data		



  cache_mem cache_mem1(.*);
  
  
 /**********  FIFO, EMULATES LEAST RECENTLY USED ENTRY **********/
  // Stores the address for later invalitae when full 
  fifo_mem LRU_fifo(.clk(clk), 
                    .rst_n(rst_n), 
                    .empty(empty), 
                    .full(full),
                    .data_out(last_entry),
                    .error(error),
                    .data_in(addr_r),
                    .push(push), 
                    .pop(pop));
  
  chk_fifo chk_fifo_1(.*, 
                      .data_out(last_entry), 
                      .data_in(addr_r),
                      .buffer(LRU_fifo.buffer),
                      .wr_addr(LRU_fifo.wr_addr), 
                      .rd_addr(LRU_fifo.rd_addr)); 
  
endmodule : cache_controller
























