/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
// -------------------------------------------------- 
// PROPERTY MODULE for FIFO
// This module is used for verification of the FIFO, and is 
// intended to be bound (with the SystemVerilog "bind") to the DUV
//import cache_ctl_pkg::*;
checker  chk_fifo (
  event clk, logic rst_n, 
  logic empty, full,
  word_t data_out,
  logic error,
  word_t data_in,
  logic push, pop,
  buffer_t buffer, // fifo storage
  logic [BIT_DEPTH : 0] wr_addr, // write fifo address  
  logic [BIT_DEPTH : 0] rd_addr); // read fifo address 
  timeunit 1ns;   timeprecision 100ps;
  default clocking default_clk @clk; endclocking
  // default disable iff !rst_n;  
// =====
 // Push on full with no pop error 
  //ap_push_error_1 : assert property (p_push_error);
  mp_push_error : assume property (not (push && full && !pop)); 
  
  // No pop on empty 
  ap_pop_error_1 : assert property (not (pop && empty));

  // full and a reset 
  cp_t1_full_1: cover property (full ##1 rst_n==0);
  
  cp_t3_empty_1: cover property (empty ##1 rst_n==0);
  
// turned unbounded into bounded
  // coverage of sequences 
  cp_push_pop_sequencing  : cover property  (push |-> ##[1:99] pop);

 
// ============= Interface Properties:

// local variables for verification purpose
// define status

function int fill_level;
fill_level  = (wr_addr[BIT_DEPTH] != rd_addr[BIT_DEPTH]); 
endfunction

function fifo_is_empty_fn;
fifo_is_empty_fn  = (wr_addr == rd_addr);
endfunction

function fifo_is_full_fn;
fifo_is_full_fn  = (wr_addr[BIT_DEPTH] != rd_addr[BIT_DEPTH]);  
endfunction



 // Definition of a fifo_b.full fifo, based on environment addresses.
  ap_fifo_full : assert property (full |-> (wr_addr[BIT_DEPTH] != rd_addr[BIT_DEPTH]));
  // If empty fifo, check empty flag   
  ap_fifo_empty : assert property (empty |-> wr_addr == rd_addr);
 property p_push_data; 
  word_t v_data;
  logic [BIT_DEPTH-1 : 0] v_ptr;
     (push, v_data = data_in, v_ptr = wr_addr) |=> buffer[v_ptr] == v_data; 
  endproperty : p_push_data
  ap_push_data : assert property (p_push_data);

  ap_pop_data : assert property (pop |-> data_out == buffer[rd_addr]);

//===== correct pointer manuipulation:
  property p_push; 
    @ (posedge clk)  disable iff (!rst_n) 
	push |=> wr_addr == ($past(wr_addr) + 1'b1);  
  endproperty : p_push
  ap_push : assert property (p_push);

  property p_pop; 
    @ (posedge clk)  disable iff (!rst_n) 
	pop  |=> rd_addr == ($past(rd_addr) + 1'b1);  
  endproperty : p_pop
  ap_pop: assert property (p_pop);

 property p_nopush; 
    @ (posedge clk)  disable iff (!rst_n) 
	!push |=> wr_addr == $past(wr_addr);  
  endproperty : p_nopush
  ap_nopush : assert property (p_nopush);

  property p_nopop; 
    @ (posedge clk)  disable iff (!rst_n) 
	!pop  |=> rd_addr == $past(rd_addr);  
  endproperty : p_nopop
  ap_nopop: assert property (p_nopop);


endchecker  : chk_fifo

// =====
//  bind fifo chk_fifo chk_fifo_1 (..);
