/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
import cache_ctl_pkg::*;

checker chk_invalidate(
   logic [0: 2**MAIN_MEM_ADDRESS_WIDTH-1] cache_valid, 
   logic [MAIN_MEM_ADDRESS_WIDTH-1 : 0] [0 : (cache_ctl_pkg::CACHE_SIZE-1)] cache_mem_addr, 
   logic [CACHE_SIZE-1: 0]  cache_entry, 
   logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr2invalidate, // address to check
   logic found_existing_entry, 					   
   event clk, logic rst_n);
  timeunit 1ns;   timeprecision 100ps;

  if (CACHE_SIZE > 1024) $error("cache size is too large"); 
  default clocking default_clk @clk; endclocking
  default disable iff !rst_n;

  // Function to search through the cache and check for the desired entry 
   function logic check_cache_entry4MT();
    automatic logic success   =0; // If==1 then it was not invalidated
    automatic int   i;
     for (i=0; i <= CACHE_SIZE -1; i++) begin : for1
       if(cache_entry[i]==1 && cache_mem_addr[i] == addr2invalidate) begin : if_1
         success  = 1; // found cache line
         break;
       end : if_1
     end : for1
    if (success) check_cache_entry4MT  =0; // was not invalidated
    else check_cache_entry4MT	 =1; // was invalidated
   endfunction : check_cache_entry4MT 
  // Use this assertion as a concurrent assertion to check for an invalidate 
  ap_invalidate : assert property(
	  ##1 cache_valid[addr2invalidate] ==0 && check_cache_entry4MT());

  //Use this assertion to warn that there is nothing to invalidate
  ap_nothing2invalidate: assert property (found_existing_entry) 
      else $error("nothing to invalidate");


endchecker : chk_invalidate

