/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
checker chk_rd_cntrl(
     //user interface
      logic data_vld, busy,  
      logic rd, full, push, pop, 
      logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr,                     
      logic [MEM_DATA_WIDTH-1:0]         rd_data, 
      // main memory interface 
      logic [MAIN_MEM_ADDRESS_WIDTH-1:0] mem_addr, 
      logic  [MEM_DATA_WIDTH-1:0]         mem_rd_data, 
      // Signals needed by checker                        
      logic cache_exists, miss,  // exists at specified address
      logic [MEM_DATA_WIDTH-1:0]         cache_data,
	  logic [2**MAIN_MEM_ADDRESS_WIDTH-1:0]  cache_valid, // .cache_mem1.cache_valid
      // clock and resets 
      event clk, 
      logic rst_n);

  timeunit 1ns;     timeprecision 100ps;
   default clocking default_clk @clk; endclocking
  
    property p_read_miss;
	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
 	  (rd && miss, v_addr=addr) |-> ##MEM_CYCLES data_vld ##1 cache_valid[v_addr];
	endproperty : p_read_miss
	ap_write2cache : assert property(p_read_miss);

    // if read and hit, thatn rd_data== cache_data
    ap_read_hit : assert property(
       rd && cache_exists |=> data_vld && rd_data==cache_data);

    // cover rd and miss
    cp_rd_miss: cover property(rd && miss);

    // cover rd and hit
    cp_rd_hit: cover property(rd && cache_exists);
  

   // Verify that if a Cache miss, then the correct data is fetched  and outputted
  	  property p_miss_then_data_2output;
		logic [MEM_DATA_WIDTH-1:0] 		 v_data;
		logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
        (rd && miss, v_addr=addr) |-> 
            ##MEM_CYCLES (data_vld && mem_addr==v_addr, v_data=mem_rd_data) 
                 |-> rd_data==v_data;
	  endproperty : p_miss_then_data_2output
		ap_miss_then_data_2output : assert property(
              p_miss_then_data_2output);
    // --------------------------------------------------
	// Verify the push and pop operations
	  ap_fifo_pop_on_miss_full: assert property(rd && miss && full |=> pop);
      ap_fifo_push_on_miss: assert property(rd && miss |=> push);

     
    // --------------------------------------------------		  		
	// Verify that for a user interface read and miss,
    //  busy is asserted for 3 cycles, and at the 3rd cycle data_vld==1
		 ap_read_miss_busy_data_vld: assert property(
				 (rd && miss) |=> busy[*MEM_CYCLES] intersect 
                                   (!data_vld[*MEM_CYCLES-1] ##1 data_vld));
   

endchecker : chk_rd_cntrl
  ////////////////////////////////////////////////////////////		





















