/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
checker chk_wr_cntrl(
     //user interface
      logic busy,  
      logic rd, wr, 
      logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr, wr_addr,                     
	  logic  [MEM_DATA_WIDTH-1:0]         wr_data, 
      // main memory interface 
      logic [MAIN_MEM_ADDRESS_WIDTH-1:0] mem_addr, 
      logic [MEM_DATA_WIDTH-1:0]         mem_wr_data, 
      logic mem_wr, 						
      // Signals needed by checker       
	  logic wr_cache,  // instructs cache to write a line //                  
	  logic [MEM_DATA_WIDTH-1:0]         cache_wr_data, // data to write to cache
	  logic [2**MAIN_MEM_ADDRESS_WIDTH-1:0]  cache_valid, // .cache_mem1.cache_valid
	  logic [(CACHE_SIZE-1):0] [MEM_DATA_WIDTH-1:0]  cache_mem_data, // cache data storage 
      logic [MAIN_MEM_ADDRESS_WIDTH-1 : 0] [0 : (CACHE_SIZE-1)] cache_mem_addr, 
      logic [CACHE_SIZE-1: 0]  cache_entry, 
      // clock and resets 
      event clk, 
      logic rst_n);				 

  timeunit 1ns;   timeprecision 100ps;
 //  This checker checks that main memory and the cache are updated for every user write 
 // and that minimum access requirements are met (BEN this can be an instantiation 
 // of the chk_min_access checker if desired). Since the memory is parameterized, 
  // elaboration time checks are used to confirm that the parameter value has integrity. 
  // This checker will be placed statically in the code.
default clocking default_clk @clk; endclocking  
	// 2. If a write from user interface, then write must be forwarded 
    // to main memory and to the cache.
	//** PROCESS: Writing these properties into the one location
	// clearly separates the assertions from the design.
	// Ideas about requirements flow in, with the freedom to
	// add supporting code without worry about interference
	// or synthesis issues of main code.

    // Verify that if a Cache write, then data is in the cache
  function logic[CACHE_SIZE-1 :0] get_index();
	//logic success=0;
	for (int j=0; j<= CACHE_SIZE-1; j=j+1) begin : for0
        // $display("infinite loop here ???  L65");             
        if(cache_entry[j]==1 && cache_mem_addr[j]==wr_addr) begin : if1// found line
//          found_existing_entry = 1'b1;
          get_index = j;
          //success =1;
		  // $display("L68 wr_cache existing no invalidate, found_index=%d", found_index);
		  break;
          // $display("found existing entry in wr L70");
        end : if1 
    end : for0
  endfunction : get_index

   property p_write2cache;
	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
	  logic [MEM_DATA_WIDTH-1:0] 		 v_wr_data;
      int   v_index;
 	   (wr_cache, v_addr=wr_addr, v_wr_data=cache_wr_data, v_index=get_index()) |=>
			cache_mem1.cache_valid[v_addr] && cache_mem1.cache_mem_addr[v_index] == v_addr && 
            cache_mem1.cache_mem_data[v_index] == v_wr_data;
  endproperty : p_write2cache
  ap_write2cache : assert property(p_write2cache);

//    property p_write2cache;  // <<< ERROR >> 
//	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
//	  logic [MEM_DATA_WIDTH-1:0] 		 v_wr_data;
// 	   (wr_cache, v_addr=wr_addr, v_wr_data=cache_wr_data) |=>
//			cache_valid[v_addr] && cache_mem_addr[get_index()] == v_addr && 
//            cache_mem_data[get_index()] == v_wr_data;
//	endproperty : p_write2cache
//	ap_write2cache : assert property(p_write2cache);

	// 1. Main memory is slow, and requires MEM_CYCLES cycles to read or to write.
	// Address and data and control signals (read and write) must be stable for 3 cycles
	// Verify that if a write, then data flows into the main memory. 
	  ap_write2main_memwr_busy : assert property(
               wr |=> mem_wr[*MEM_CYCLES] intersect  busy[*MEM_CYCLES]);

      property p_write2main_memaddr_memwrdata;
	    logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
	    logic [MEM_DATA_WIDTH-1:0] 		   v_wr_data;
		(wr, v_addr=addr,  v_wr_data=wr_data) |=> 
                 mem_addr==v_addr[*MEM_CYCLES] intersect mem_wr_data==v_wr_data[*MEM_CYCLES];
	  endproperty : p_write2main_memaddr_memwrdata
	   ap_write2main : assert property(p_write2main_memaddr_memwrdata);

	// Verify that for a user interface wr, the wr signal is one cycle, and rd is off
	   ap_write_controls :assert property(
               wr |=> !wr[*MEM_CYCLES] and !rd[*MEM_CYCLES] and busy[*MEM_CYCLES]);

	 // If busy, then no rd/wr  [Ben] FIXED 
       always @ (posedge clk)
		 ap_busy_no_rd_wr: assert property (busy |-> !rd && !wr);

	 // If rd==1 && wr==1'b0 and in cache, 
	 // then @ next cycle mem_wr==1'b0, mem_rd==1'b0, data_vld==1b1, busy==1'b0,
	 // rd_data== the cache data   

	  // never a simulataneous rd and wr
       always @ (posedge clk) 
		 ap_never_rd_wr: assert (! (rd && wr));

endchecker : chk_wr_cntrl
  ////////////////////////////////////////////////////////////		

	

	  



// This checker checks that when a miss occurs, the data is fetched from main memory 
// and the data is provided back to the user and replaces the oldest data in the cache  
// (BEN - might want to inject an error of forgetting to send that data back to the user.)
//  The data_vld signal is asserted three rising clocks after the rd signal is asserted. 
// This checker will be placed procedurally in the code branch that addresses misses.


  


  // 4. If a read from the user interface is requested and the data is in the cache, 
  // then the data will supplied to the user interface at the next cycle.  
  // Otherwise, the data will be fetched from the main memory, stored into the cache,
  //  and be available as valid (with data_vld ==1'b1) at the 3rd cycle after the read request. 

  // CODE + checker 


/* -----\/----- EXCLUDED -----\/-----


chk_min_access:  This checker provides assertions that will flag when minimum access times are violated. The checker will be used four times, once for the read and the write interface for both main memory and for cache memory. (Ben - this can be in the form of a checker in a package that has formal arguments with the inferred clock and disable.  We could add a global clock to ensure that data is stable but we'd have to define the requirement )
 
chk_user_writes:  This checker checks that main memory and the cache are updated for every user write and that minimum access requirements are met (BEN - this can be an instantiation of the chk_min_access checker if desired). Since the memory is parameterized, elaboration time checks are used to confirm that the parameter value has integrity. This checker will be placed statically in the code.
 
chk_cache_misses:  This checker checks that when a miss occurs, the data is fetched from main memory and the data is provided back to the user and replaces the oldest data in the cache  (BEN - might want to inject an error of forgetting to send that data back to the user.). The data_vld signal is asserted three rising clocks after the rd signal is asserted. This checker will be placed procedurally in the code branch that addresses misses.
 
chk_cache_hits:  This checker checks that when a hit (!miss) occurs the proper data is returned to the user and main memory is not accessed.  The data_vld signal is asserted on the next rising edge of the clk. This checker will be placed procedurally in the code branch that addresses hits. 
 
chk_cache_invalidate: When the cache is full and more data needs to be added, this checker checks that the oldest data is invalidated and then overwritten.  (BEN:  it is not clear to me how you do this.  How is the oldest data identified and why doesn't the overwrite serve the invalidate purpose).  If invalidate is needed, then you need to also check that cache hits do not reflect invalidated data)
 
chk_infc_integrity: This checker checks that reads and writes of data to the various memories do not collide and that control and data signals are never X or Z values. 
 

In addition to the verification checks, cover statements will be added to the checkers.

 -----/\----- EXCLUDED -----/\----- */

  /////////////////////////////////////////////////////////
  
